/*******************************************************************************
 * Copyright (c) 2016 Dec. Verimag. All rights reserved.
 * @author Hang YU
*******************************************************************************/

#include <iostream>
#ifndef _GLPKINTER
#define _GLPKINTER
#include "glpkInterface.h"
#endif
#include "raytracing.h"

Polyhedron::Polyhedron (int consNum, int variNum) 
  : _constraints_num(consNum), _variable_num(variNum), _is_empty(true) {
  _coefficients.resize(consNum, variNum) ;
  _constants.resize(consNum) ;
  _active_table.resize(consNum, false) ;
}

/*******************************************************************************
 * Calls the raytracing method to eliminate the redundant constraints, and sets
 * the corresponding value in _active_table as false
*******************************************************************************/
void Polyhedron::Minimize (int method) {
  Raytracing raytrace( *this, _central_point ) ;
  
  time_t begin = clock() ;
  switch (method) { 
    case 1: 
      raytrace.RayHitting() ; break ;
    case 2: 
      raytrace.RayHittingTwoDirect() ; break ;
    case 3: 
      raytrace.RayHittingMatrix() ; break ;
    case 4:
      raytrace.RayHittingMatrixTwoDir() ; break ;
    default:
      std::cout << "Illegal parameter of Minimize()" << std::endl ;
      abort() ;
  }  
  time_t end = clock() ;
  double runtime = (end - begin) * 1000 / CLOCKS_PER_SEC ;
  std::cout << "RayHitting() time: " << runtime << " ms" << std::endl ;

  //PrintActiveIdx() ;

  raytrace.Determine() ;
}

int Polyhedron::get_variable_num () const {
  return _variable_num ;
}

void Polyhedron::set_variable_num (int num) {
  _variable_num = num ;
}

int Polyhedron::get_constraints_num () const {
  return _constraints_num ;
}

void Polyhedron::set_constraints_num (int num) {
  _constraints_num = num ;  
}

void Polyhedron::Activate (int consIdx) {
  _active_table[consIdx] = true ;
}

void Polyhedron::Deactivate (int consIdx) {
  _active_table[consIdx] = false ;
}

bool Polyhedron::IsActive (int consIdx) const {
  return _active_table[consIdx] ;
}

void Polyhedron::set_central_point(const Point& point) {
  _central_point = point ;
  _is_empty = false ;
}

Point Polyhedron::get_central_point () {
  if ( _central_point.IsEmpty() ) {
    GlpkInterface glpkinter ;
    _central_point = glpkinter.GetCentralPoint(*this) ;
  }
  return _central_point ;
}

Vector Polyhedron::GetConstraint (int consIdx) const {
  return _coefficients.row(consIdx) ;
}

Matrix Polyhedron::get_coefficients () const {
  return _coefficients ;
}
 
void Polyhedron::SetConstraint (int consIdx, const Vector& cons) {
  _coefficients.row(consIdx) = cons ;
}

double Polyhedron::GetConstant (int consIdx) const {
  return _constants(consIdx) ;
} 

void Polyhedron::SetConstant (int consIdx, double val) {
  _constants(consIdx) = val ;
}

double Polyhedron::GetCoef (int consIdx, int variIdx) const {
  return _coefficients(consIdx, variIdx) ;
}

void Polyhedron::SetCoef (int consIdx, int variIdx, double val) {
  _coefficients(consIdx, variIdx) = val ;
}


/*******************************************************************************
 * Prints the index of the avtivate constraints
*******************************************************************************/
void Polyhedron::PrintActiveIdx () {
  std::cout<< "The index of the activate constraints are: " << std::endl ;
  for (int i = 0; i < (int)_active_table.size(); ++ i) {
    if (_active_table[i] == 1) {
      std::cout << i+1 << " " ;
    } 
  }
  std::cout << std::endl ;
}

/*******************************************************************************
 * Prints the constraints to the terminal
*******************************************************************************/
void Polyhedron::Print () {
  std::cout << "The constraints are:" << std::endl ;
  for (int i = 0; i < _constraints_num; ++ i) {
    for (int j = 0; j < _variable_num; ++ j) {
      std::cout << _coefficients(i, j) << " " ;
    }
    std::cout << " " << _constants[i] << std::endl ;
  }
}

// TODO
/*
int Polyhedron::AddConstraint (const Vector& constraint) {

}

bool Polyhedron::Satisfy (const Vector& constraint) {

}

bool Polyhedron::IsSatisfiedBy (const Point& point) {

}
    
bool Polyhedron::IsConstraintActive (int i) {
  return _active_table[i] ;  
}
  
double Polyhedron::Evaluate (const Point& point, int consIdx) {

}
  
bool Polyhedron::IsIncludedIn (const Vector& constraint) {

}

bool Polyhedron::IsEmpty () const {
  return _is_empty ;
}
 
void Polyhedron::Print () {

}
  
void Polyhedron::Print (int consIdx) {

}
  
void Polyhedron::PrintSage () {

}
  
void Polyhedron::PrintSage (int consIdx) {

}
*/
