/*******************************************************************************
 * Copyright (c) 2016 Dec. Verimag. All rights reserved.
 * @author Hang YU
*******************************************************************************/

#include <iostream>
#ifndef _GLPKINTER
#define _GLPKINTER
#include "glpkInterface.h"
#endif
#include "raytracing.h"
#include "apronInterface.h"
#ifndef _DOUBLE
#define _DOUBLE
#include "double.h"
#endif

Polyhedron::Polyhedron (int consNum, int variNum) 
  : _constraints_num(consNum), _variable_num(variNum), _is_minimized(false) {
  _coefficients.resize(consNum, variNum) ;
  _constants.resize(consNum) ;
  _active_table.resize(consNum, false) ;
}

/*******************************************************************************
 * Calls the raytracing method to eliminate the redundant constraints, and sets
 * the corresponding value in _active_table as false
*******************************************************************************/
void Polyhedron::Minimize (int method) {
  // eliminate the duplicate constraints
  Merge() ;
  if ( IsEmpty() ) {
    std::cerr << "Minimization failed. The polyhedron is empty." << std::endl ;
    std::terminate() ;
  }
  if (_is_minimized) {
    return ;
  }
  if ( _central_point.IsEmpty() ) {
    set_central_point( GlpkInterface::GetCentralPoint(*this) ) ;
    if(_central_point.IsEmpty()) {
      std::cerr << "Minimization failed. Cannot find a point inside the polyhedra" << std::endl ;
      std::terminate() ;
    }
  }
  Raytracing raytrace(*this, _central_point) ;
  switch (method) { 
    case 1: 
      raytrace.RayHitting() ; break ;
    case 2: 
      raytrace.RayHittingTwoDirect() ; break ;
    case 3: 
      raytrace.RayHittingMatrix() ; break ;
    case 4:
      raytrace.RayHittingMatrixTwoDir() ; break ;
    default:
      std::cerr << "Illegal parameter of Minimize()" << std::endl ;
      std::terminate() ;
  }  
  raytrace.Determine() ;
  _is_minimized = true ;
}
/*******************************************************************************
 * Minimization without raytracing.
*******************************************************************************/
void Polyhedron::MinimizeSimple () {
  Merge() ;
  std::vector<bool> res(_constraints_num) ;
  // #pragma omp parallel for
  for (int i = 0; i < _constraints_num; ++ i) {
    res[i] = GlpkInterface::Sat(*this, i) ;
    if (res[i] == true) {
      Activate(i) ;       
    }
  }  

  _is_minimized = true ;
}

/*******************************************************************************
 * Gets the minimized polyhedron. If the polyhedron is not minimized, minimize
 * it first.
 * @return the minimized polyhedron
*******************************************************************************/
Polyhedron Polyhedron::GetMinimizedPoly () {
  if (! _is_minimized) {
    Minimize() ;
  }
  std::vector<int> indexVec ;
  for (int i = 0; i < (int)_active_table.size(); ++ i) {
    if (_active_table[i] == true) {
      indexVec.push_back(i) ;
    }
  }
  Polyhedron miniPoly = GetSubPoly(indexVec) ;

  return miniPoly ;
}

/*******************************************************************************
 * Compares with the result of minimization by Apron.
 * @return true if the results are equal
*******************************************************************************/
bool Polyhedron::EqualApronMinimize () {
  ApronInterface apronObj(*this) ;
  Polyhedron miniPoly = GetMinimizedPoly() ; 
  ApronInterface apronMiniObj(miniPoly) ;
  apronObj.Print() ;
  apronMiniObj.Print() ;

  return apronObj.CompareAbstract(apronMiniObj) ;
}

/*******************************************************************************
 * Creates a polyhedron which contains a subset of constraints of the current
 * polyhedron.
 * @para index the index of the constraints in the subset
 * @return the new polyhedron contains the subset of constraints 
*******************************************************************************/
Polyhedron Polyhedron::GetSubPoly (const std::vector<int>& indexVec) {
  int consNum = indexVec.size() ;
  int variNum = get_variable_num() ;
  Polyhedron newPoly(consNum, variNum) ;
  int currIdx ;
  for (int i = 0; i < consNum; ++ i) {
    currIdx = indexVec[i] ;
    for (int j = 0; j < variNum; ++ j) {
      newPoly.SetCoef( i, j, GetCoef(currIdx, j) ) ;
    }
    newPoly.SetConstant( i, GetConstant(currIdx) ) ;
  }
  return newPoly ;
}

/*******************************************************************************
 * Checks inclusion with raytracing. 
 * For example we check if p1 is included in P2, we need three steps:
 * 1) Check where is the inner point x of P1. If x is not inside P2, then P1
 * is NOT included in P2; 
 * 2) Check each constraint in P2. If any constraint in P2 is irredundant wrt P1
 * then P1 is not included in P2;
 * 3) If no constraint in P2 is irredundant wrt P1 then P1 is included in P2.
 * Note that P1 and P2 should be minimized.
 * @para poly the polyhedron which may be included in the current one, i.e. here
 * we check if poly is included in *this.
 * @return true if poly is included in the current polyhedron 
*******************************************************************************/
bool Polyhedron::Include (const Polyhedron& poly) {
  if ( get_variable_num() != poly.get_variable_num() ) {
    std::cerr << "Caonnt check inclusion. Polyhedra are not in the same environment." 
      << std::endl ;
    std::terminate() ;
  }
  const int consNumP1 = poly.get_constraints_num() ;
  const int consNumP2 = get_constraints_num() ;
  // check if x is inside P2
  for (int i = 0; i < consNumP2; ++ i) {
    if (Satisfy(poly.get_central_point(), i) == false) {
      return false ;
    }
  } 
  // No need to call Merge() here, as duplicate constraints are treated as redundant 
  Polyhedron newPoly = Combine(poly) ;
  // Need to use the central point of the new polyhedron to do raytracing, otherwise
  // the central point may be too closed to some constraints and cause raytracing
  // cannot determine if the constraints are redundant.
  newPoly.set_central_point( GlpkInterface::GetCentralPoint(newPoly) ) ;
  if(newPoly._central_point.IsEmpty()) {
    std::cerr << "Cannot check inclusion. Cannot find a point inside the polyhedra" << std::endl ;
    std::terminate() ;
  }
  Raytracing raytrace( newPoly, newPoly._central_point ) ;
  // check constraints of P2  
  raytrace.RayHittingMatrixTwoDir(true, consNumP1) ;    
  if (raytrace.HasInclusion() == false) {
    return false ;
  }
  raytrace.Determine(true, consNumP1) ;
  if (raytrace.HasInclusion() == false) {
    return false ;
  }

  return true ;
}

/*******************************************************************************
 * Put constraints in two polyhedra together in a new polyhedron. The polyhedron
 * at the parameter is put at the beginning.
 * @para poly the polyhedron to be put at the beginning
 * @return the new polyhedron contains all constraints 
*******************************************************************************/
Polyhedron Polyhedron::Combine (const Polyhedron& poly) {
  int consNum1 = poly.get_constraints_num() ;
  int consNum2 = get_constraints_num() ;
  int variNum = get_variable_num() ;
  Polyhedron newPoly(consNum1+consNum2, variNum) ;
  for (int i = 0; i < consNum1; ++ i) {
    for (int j = 0; j < variNum; ++ j) {
      newPoly.SetCoef( i, j, poly.GetCoef(i,j) ) ;
    }
    newPoly.SetConstant( i, poly.GetConstant(i) ) ;
  }
  for (int i = 0; i < consNum2; ++ i) {
    for (int j = 0; j < variNum; ++ j) {
      newPoly.SetCoef( consNum1+i, j, GetCoef(i,j) ) ;
    }
    newPoly.SetConstant( consNum1+i, GetConstant(i) ) ;
  }

  return newPoly ;
}

/*******************************************************************************
 * Checks if a constraint satisfy a point.
 * @para point the point to be satisfied 
 * @para index the index of the constraint to be checked
 * @return true if the constraint satisfy the point
*******************************************************************************/
bool Polyhedron::Satisfy (const Point& point, int index) {
  double res = point.get_coordinates().dot( GetConstraint(index) ) ; 
  if ( Double::IsLessEq(res, GetConstant(index)) ) {
    return true ;
  }
  return false ;
}

/*******************************************************************************
 * Merge the same constraints
*******************************************************************************/
void Polyhedron::Merge () {
  std::vector<int> idx ;
  bool dup = false ;
  idx.push_back(0) ;
  for (int i = 1; i < _constraints_num; ++ i) {
    for (int k = i-1; k >= 0; -- k) {
      dup = ConstraintsEq(i, k) ;
      if (dup == true) break ;
    }
    if (dup == false) {
      idx.push_back(i) ;
    } 
  }
  *this = GetSubPoly(idx) ;
}
  
/*******************************************************************************
 * @para idx1 the index of the first constraint
 * @para idx2 the index of the second constraint
 * @return true if two constraints are equal
*******************************************************************************/
bool Polyhedron::ConstraintsEq (int idx1, int idx2) {
  int variIdx2 = -1 ;
  // find the first not zero coefficient
  for (int j = 0; j < _variable_num; ++ j) {
    if (variIdx2 == -1 && _coefficients(idx2, j) != 0) {
      variIdx2 = j ;
      break ;
    }
  }
  double temp12 = _coefficients(idx1, variIdx2) ; // may be zero
  double temp22 = _coefficients(idx2, variIdx2) ; // not zero
  // double check
  double lamda = temp12 / temp22 ; 
  for (int j = 0; j < _variable_num; ++ j) {
    if (! Double::AreEqual(_coefficients(idx1, j), _coefficients(idx2, j) * lamda) ) {
      return false ;
    }
  }
  if (! Double::AreEqual( _constants(idx1), _constants(idx2) * lamda) ) {
    return false ;
  }

  return true ;
}

int Polyhedron::get_variable_num () const {
  return _variable_num ;
}

void Polyhedron::set_variable_num (int num) {
  _variable_num = num ;
}

int Polyhedron::get_constraints_num () const {
  return _constraints_num ;
}

void Polyhedron::set_constraints_num (int num) {
  _constraints_num = num ;  
}

void Polyhedron::Activate (int consIdx) {
  _active_table[consIdx] = true ;
}

void Polyhedron::Deactivate (int consIdx) {
  _active_table[consIdx] = false ;
}

bool Polyhedron::IsActive (int consIdx) const {
  return _active_table[consIdx] ;
}

void Polyhedron::set_central_point(const Point& point) {
  _central_point = point ;
}

Point Polyhedron::get_central_point () const {
  return _central_point ;
}

Vector Polyhedron::GetConstraint (int consIdx) const {
  return _coefficients.row(consIdx) ;
}

Matrix Polyhedron::get_coefficients () const {
  return _coefficients ;
}
 
void Polyhedron::SetConstraint (int consIdx, const Vector& cons) {
  _coefficients.row(consIdx) = cons ;
}

double Polyhedron::GetConstant (int consIdx) const {
  return _constants(consIdx) ;
} 

void Polyhedron::SetConstant (int consIdx, double val) {
  _constants(consIdx) = val ;
}

double Polyhedron::GetCoef (int consIdx, int variIdx) const {
  return _coefficients(consIdx, variIdx) ;
}

void Polyhedron::SetCoef (int consIdx, int variIdx, double val) {
  _coefficients(consIdx, variIdx) = val ;
}


/*******************************************************************************
 * @return true if the polyhedron is empty
*******************************************************************************/
bool Polyhedron::IsEmpty () const {
  if (_coefficients.size() == 0)
    return true ;
  return false ;
}
  
/*******************************************************************************
 * @return true if the polyhedron is minimized
*******************************************************************************/
bool Polyhedron::IsMinimized () const {
  return _is_minimized ;
}

/*******************************************************************************
 * Prints the index of the avtivate constraints
*******************************************************************************/
void Polyhedron::PrintActiveIdx () const {
  std::cout<< "The index of the activate constraints are: " << std::endl ;
  for (int i = 0; i < (int)_active_table.size(); ++ i) {
    if (_active_table[i] == 1) {
      std::cout << i << " " ;
    } 
  }
  std::cout << std::endl ;
}

/*******************************************************************************
 * Prints the constraints to the terminal
*******************************************************************************/
void Polyhedron::Print () const {
  std::cout << "The constraints are (in form [coefficients | constant]):" << std::endl ;
  for (int i = 0; i < _constraints_num; ++ i) {
    for (int j = 0; j < _variable_num; ++ j) {
      std::cout << _coefficients(i, j) << " " ;
    }
    std::cout << " " << _constants[i] << std::endl ;
  }
}

 


// TODO
/*
int Polyhedron::AddConstraint (const Vector& constraint) {

}

bool Polyhedron::Satisfy (const Vector& constraint) {

}

bool Polyhedron::IsSatisfiedBy (const Point& point) {

}
    
bool Polyhedron::IsConstraintActive (int i) {
  return _active_table[i] ;  
}
  
double Polyhedron::Evaluate (const Point& point, int consIdx) {

}
  
bool Polyhedron::IsIncludedIn (const Vector& constraint) {

}
 
void Polyhedron::Print () {

}
  
void Polyhedron::Print (int consIdx) {

}
  
void Polyhedron::PrintSage () {

}
  
void Polyhedron::PrintSage (int consIdx) {

}
*/
